from tqdm import tqdm
import gzip
import json
import copy
import os
import sys
import time
import subprocess
import click
import urllib
import tarfile
from pathlib import Path

import_tweets = """
    WITH {json} as data
    UNWIND data AS row
    MERGE (r:Tweet{id: row.id}) 
    SET r+=row, r.created_at=datetime(row.created_at), r.recorded_at=datetime(row.recorded_at)
    """

import_users = """
    WITH {json} as data
    UNWIND data AS row
    MERGE (r:User{id: row.id}) 
    SET r+=row, r.created_at=datetime(row.created_at), r.recorded_at=datetime(row.recorded_at)
    """

merge_posts = """
    WITH {json} as data
    UNWIND data AS row
    MATCH (t:Tweet{id: row.tweet_id})
    MATCH (u:User{id: row.user_id})
    MERGE (u) -[:POSTS]-> (t)
    """

merge_refers = """
    WITH {json} as data
    UNWIND data AS row
    MATCH (t:Tweet{id: row.tweet_id})
    MATCH (r:Tweet{id: row.ref_tweet_id})
    MERGE (t) -[:REFERS_TO]-> (r)
    """

merge_mentions = """
    WITH {json} AS data
    UNWIND data AS row
    MATCH (t:Tweet{id: row.tweet_id})
    MATCH (u:User{id: row.mentioned_id})
    MERGE (t) -[:MENTIONS]-> (u)
    """

create_tags = """
    WITH {json} AS data
    UNWIND data as tag
    MERGE (t:Tag{id:tag})
    """

merge_tags = """
    WITH {json} AS data
    UNWIND data as row
    MATCH (t:Tweet{id: row.tweet_id})
    MATCH (r:Tag{id: row.tag})
    MERGE (t) -[:HAS_TAG]-> (r)
    """

create_urls = """
    WITH {json} AS data
    UNWIND data as url
    MERGE (t:Url{id:url})
    """

merge_urls = """
    WITH {json} AS data
    UNWIND data as row
    MATCH (t:Tweet{id: row.tweet_id})
    MATCH (u:Url{id: row.url})
    MERGE (t) -[:HAS_URL]-> (u)
    """

def import_records(graph, records):

    # Generate indexes (if not exist)
    graph.run("CREATE CONSTRAINT ON (r:Tweet) ASSERT r.id IS UNIQUE")
    graph.run("CREATE INDEX ON :Tweet(created_at)")
    graph.run("CREATE INDEX ON :Tweet(type)")
    graph.run("CREATE CONSTRAINT ON (r:User) ASSERT r.id IS UNIQUE")
    graph.run("CREATE INDEX ON :User(created_at)")
    graph.run("CREATE INDEX ON :User(screen_name)")
    graph.run("CREATE CONSTRAINT ON (t:Tag) ASSERT t.id IS UNIQUE")
    graph.run("CREATE CONSTRAINT ON (u:Url) ASSERT u.id IS UNIQUE")
    graph.sync()

    q = "CALL db.indexes() YIELD indexName AS i, type AS t WHERE t = 'node_fulltext' RETURN collect(i)"
    existing = graph.run(q).single().value()
    if 'tweets' not in existing:
        graph.run("CALL db.index.fulltext.createNodeIndex('tweets', ['Tweet'], ['text'])")
        print("Created fulltext index for tweets")
    if 'users' not in existing:
        graph.run("CALL db.index.fulltext.createNodeIndex('users', ['User'], ['screen_name', 'description'])")
        print("Created fulltext index for users")
    graph.sync()

    # Import recordings (tweets and users)
    seen_tags = set()
    seen_urls = set()
    imported = [];
    if Path('imported.json').exists():
        with open('imported.json') as f:
            imported = json.load(f)
    
    tbd = sorted(list(set(records) - set(imported)))
    p = tqdm(tbd, desc="Importing", file=sys.stdout)
    for f in p:

        with gzip.open(f, "rb") as chunk:
            data = json.loads(chunk.read().decode("utf-8"))
            
            # Import users
            users = [d for d in data if d['type'] == 'user']
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(users)} users)")
            graph.run(import_users, json=users)
            
            # Import tweets
            tweets = [d for d in data if d['type'] != 'user']
            imports = copy.deepcopy(tweets)
            for t in imports:
                t.pop('mentions', None)
                t.pop('mentioned_ids', None)
                t.pop('hashtags', None)
                t.pop('urls', None)
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(tweets)} tweets)")
            graph.run(import_tweets, json=imports)
            graph.sync()
            
            # Generate post relations
            posts = [{ 'user_id': t['user'], 'tweet_id': t['id'] } for t in tweets if 'user' in t]
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(posts)} posts)")
            graph.run(merge_posts, json=posts)

            # Generate refers_to relations
            refers = [{ 'tweet_id': t['id'], 'ref_tweet_id': t['refers_to'] } for t in tweets if 'refers_to' in t]
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(refers)} referings)")
            graph.run(merge_refers, json=refers)

            # Generate mention relations
            mentions = [{ 'tweet_id': t['id'], 'mentioned_id': mid } for t in tweets for mid in t['mentioned_ids']]
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(mentions)} mentions)")
            graph.run(merge_mentions, json=mentions)

            # Generate tag relations
            tags = set([tag for t in tweets for tag in t['hashtags']]) - seen_tags
            seen_tags = seen_tags.union(tags)
            tag_rels = [{'tweet_id': t['id'], 'tag': tag.upper()} for t in tweets for tag in t['hashtags']]
            p.set_description(f"Processing {p.n}/{p.total} (merge {len(tags)} new tags)")            
            graph.run(create_tags, json=list(tags))
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(tag_rels)} taggings)")            
            graph.run(merge_tags, json=tag_rels)

            # Generate url relations
            urls = set([url for t in tweets for url in t['urls']]) - seen_urls
            seen_urls = seen_urls.union(urls)
            url_rels = [{'tweet_id': t['id'], 'url': url } for t in tweets for url in t['urls']]
            p.set_description(f"Processing {p.n}/{p.total} (merge {len(urls)} new urls)")            
            graph.run(create_urls, json=list(urls))
            p.set_description(f"Processing {p.n}/{p.total} (merging {len(url_rels)} url refers)")            
            graph.run(merge_urls, json=url_rels)
        
        imported.append(f)
        with open('imported.json', 'w') as f:
            json.dump(imported, f, indent=3)

def install_neo4j(config):
    if not Path('neo4j').exists():
        settings = json.load(open(config))
        url = "https://neo4j.com/artifact.php?name=neo4j-community-3.5.8-unix.tar.gz"
        click.echo(f"Downloading Neo4j [{ url }]")
        file, msg = urllib.request.urlretrieve(url, "neo4j.tar.gz")
        with tarfile.open(file) as tf:
            tf.extractall()
            os.system("mv neo4j*/ neo4j/")
            os.system(f"rm { file }")
            os.system(f"neo4j/bin/neo4j-admin set-initial-password '{ settings['neo4j_pwd'] }'")

def start_neo4j(config):
    if Path("neo4j").exists():
        while subprocess.getoutput("neo4j/bin/neo4j status") == "Neo4j is not running":
            click.echo("Starting Neo4j")
            os.system("neo4j/bin/neo4j start")
            time.sleep(5)

def stop_neo4j():
    if Path("neo4j").exists():
        click.echo("Stopping Neo4j")
        os.system("neo4j/bin/neo4j stop")


import nbformat as nbf

graph_connect = """
import json
from py2neo import Graph

conf = json.load(open('%s'))
graph = Graph(conf['neo4j_url'], user=conf['neo4j_usr'], password=conf['neo4j_pwd'])
""".strip()

cypher_example = """
import matplotlib.pyplot as plt

queries = [
    { 
        "title": "Distribution of node types",
        "query": "MATCH (n) RETURN labels(n) as category, count(n) AS n ORDER BY n DESCENDING LIMIT 10"
    },
    {
        "title": "Distribution of Tweet subtypes",
        "query": "MATCH (n:Tweet) RETURN n.type as category, count(n) AS n ORDER BY n DESCENDING",
    },
    {
        "title": "Distribution of edge types",
        "query": "MATCH () -[r]-> () RETURN type(r) as category, count (r) AS n ORDER BY n DESCENDING",
    },
    {
        "title": "TOP 10 of mosted retweeted users",
        "query": "MATCH (n:User) -[:POSTS]-> (t:Tweet) <-[rts:REFERS_TO]- (:Tweet{type:'retweet'}) RETURN n.screen_name AS category, count(rts) AS n ORDER BY n DESCENDING LIMIT 10"
    },
    {
        "title": "TOP 10 of mosted retweeting users",
        "query": "MATCH (n:User) -[:POSTS]-> (t:Tweet{type:'retweet'}) RETURN n.screen_name AS category, count(t) AS n ORDER BY n DESCENDING LIMIT 10"
    },
    {
        "title": "TOP 10 of most replying users",
        "query": "MATCH (u:User) -[:POSTS]-> (r:Tweet{type: 'reply'}) RETURN u.screen_name AS category, count(r) AS n ORDER BY n DESCENDING LIMIT 10"
    },
    {
        "title": "TOP 10 of most used tags",
        "query": "MATCH (t:Tweet) -[r:HAS_TAG]-> (tag:Tag) RETURN tag.id AS category, COUNT(r) AS n ORDER BY n DESCENDING LIMIT 10"
    },
    {
        "title": "TOP 10 of most retweeted urls",
        "query": "MATCH (t:Tweet{type: 'retweet'}) -[r:HAS_URL]-> (u:Url) RETURN u.id AS category, COUNT(t) AS n ORDER BY n DESCENDING LIMIT 10"
    },
    {
        "title": "TOP 10 of user locations",
        "query": "MATCH (u:User) WHERE u.location IS NOT NULL RETURN u.location AS category, count(u) AS n ORDER BY n DESCENDING LIMIT 10"
    }
]

for q in queries:
    data = graph.run(q['query']).data()
    category = [d['category'] for d in data]
    values = [d['n'] for d in data]
    plt.title(q['title'])
    plt.pie(values, labels=category, autopct='%1.1f%%', startangle=90)
    plt.axis('equal')
    plt.show()
""".strip()

def notebook(config):
    nb = nbf.v4.new_notebook()
    nb['cells'] = [
        nbf.v4.new_markdown_cell("# Graph analysis"),
        nbf.v4.new_markdown_cell(
            "Twista generated this notebook for demonstrating purposes. "
            "Copy and/or adapt it to your specfic analysis requirements. "
        ),
        nbf.v4.new_code_cell(graph_connect % config),
        nbf.v4.new_markdown_cell("## Graph query examples"),
        nbf.v4.new_markdown_cell(
            "The following queries are just some example Cypher queries "
            "demonstrating how to query the graph and how "
            "visualizing observable quantitative aspects. "
            "\n"
            "You might find the following links usefull:\n"
            "\n"
            "- [Introduction to Cypher](https://neo4j.com/developer/cypher-query-language/)\n"
            "- [Introduction to matplotlib](https://matplotlib.org/3.1.1/tutorials/introductory/pyplot.html)"
            "- [Twista Wiki]()"
        ),
        nbf.v4.new_code_cell(cypher_example)
    ]
    return nb


import json
import datetime
import twista
from dateutil import parser

class TweetObject:
    
    def __init__(self, json):
        self.json = json
        self.json['created_at'] = parser.parse(self.json['created_at']).isoformat()
        self.json['recorded_at'] = datetime.datetime.now(tz=datetime.timezone.utc).isoformat()
        
    def id(self):
        return self.json['id_str']
    
    def type(self):
        if (self.is_user()):
            return "user"
        if (self.is_status()):
            return "status" 
        if (self.is_retweet()):
            return "retweet" 
        if (self.is_quote()):
            return "quote" 
        if (self.is_reply()):
            return "reply"
        return "unknown"
    
    def created_at(self):
        return self.json['created_at']
    
    def recorded_at(self):
        return self.json['recorded_at']
    
    def user(self):
        return User(self.json['user'])
        
    def text(self):
        return self.json['text']
        
    def is_user(self):
        return "screen_name" in self.json

    def is_status(self):
        return not (self.is_quote() or self.is_reply() or self.is_retweet())
    
    def is_quote(self):
        return self.json['is_quote_status'] and "retweeted_status" not in self.json
    
    def is_reply(self):
        return self.json['in_reply_to_status_id_str'] != None
    
    def reply_to_status_id(self):
        return self.json['in_reply_to_status_id_str']
    
    def is_retweet(self):
        return "retweeted_status" in self.json
    
    def quote(self):
        if 'quoted_status' not in self.json:
            return None
        return Status(self.json['quoted_status'])
    
    def retweet(self):
        return Status(self.json['retweeted_status']) if self.is_retweet() else None
    
    def status(self):
        return Status(self.json)
    
    def __str__(self):
        return json.dumps(self.json, indent=4)
    
    
class Entities:
    
    def __init__(self, json):
        self.json = json
        
    def hashtags(self):
        return [e['text'] for e in self.json['hashtags']]

    def urls(self):
        return [e['url'] for e in self.json['urls']]
    
    def mentions(self):
        return [e['screen_name'] for e in self.json['user_mentions']]

    def mentioned_ids(self):
        res = [e['id_str'] for e in self.json['user_mentions']]
        return [r for r in res if r is not None]
    
    def as_dict(self):
        return {
            "hashtags": self.hashtags(),
            "urls": self.urls(),
            "mentions": self.mentions(),
            "mentioned_ids": self.mentioned_ids()         
        }
    
    def __str__(self):
        return json.dumps(self.as_dict(), indent=4)
        

    
class Status(TweetObject):
        
    def text(self):
        if 'extended_tweet' in self.json:
            return self.json['extended_tweet']['full_text']
        return self.json['text']
    
    def entities(self):
        if 'extended_tweet' in self.json:
            return Entities(self.json['extended_tweet']['entities'])
        return Entities(self.json['entities'])
    
    def as_dict(self):

        rts = self.json['retweet_count']
        favs = self.json['favorite_count']

        r = {
            'twista': twista.VERSION,
            "type": self.type(),
            "id": self.id(),
            "user": self.user().id(),
            "created_at": self.created_at(),
            "recorded_at": self.recorded_at(),
            "source": self.json['source'],
            "retweets": rts if rts else 0,
            "favourites": favs if favs else 0,
            "lang": self.json['lang']
        }

        r.update(self.entities().as_dict())

        if not self.is_retweet():
            r['text'] = self.text()

        if self.is_quote():
            r['refers_to'] = self.json['quoted_status_id_str']

        if self.is_retweet():
            r['refers_to'] = self.retweet().id()

        if self.is_reply():
            r['refers_to'] = self.reply_to_status_id()
            
        return r
    
    
class User(TweetObject):
                
    def screenname(self):
        return self.json['screen_name']
    
    def as_dict(self):
        return {
            'twista': twista.VERSION,
            'type': 'user',
            'id': self.id(),
            'screen_name': self.screenname(),
            'created_at': self.created_at(),
            'recorded_at': self.recorded_at(),
            'location': self.json['location'],
            'description': self.json['description'],
            'url': self.json['url'],
            'verified': self.json['verified'],
            # 'lang': self.json['lang'], deprecated meanwhile
            'followers': self.json['followers_count'],
            'friends': self.json['friends_count'],
            'listed': self.json['listed_count'],
            'favourites': self.json['favourites_count'],
            'statuses': self.json['statuses_count']
        }


import datetime
import time
import gzip
import json
import atexit
import tweepy
import traceback
from termcolor import colored
from twista.dm import TweetObject

class Recorder(tweepy.StreamListener):

    entities = {}
    N = 25000

    # Constructor 
    def __init__(self, n):
        super().__init__()
        self.N = n

    def record(self, ia):
        if ia.type() not in self.entities:
            self.entities[ia.type()] = {}
        self.entities[ia.type()][ia.id()] = ia.as_dict()

    def write(self):
        t = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H-%M")
        with gzip.open("recording-%s.json.gz" % t, 'wt') as file:
            file.write(self.as_json())
        self.entities = {}

    def length(self):
        return sum([len(d.values()) for _, d in self.entities.items()])

    def as_json(self):
        entries = []
        for _, d in self.entities.items():
            entries.extend(d.values())
        return json.dumps(entries, indent=2)

    def on_status(self, status):
        tweet = TweetObject(status._json)
        print(tweet.status().text())

        self.record(tweet.status())
        self.record(tweet.user())
        
        if tweet.is_retweet():
            self.record(tweet.retweet())
            self.record(tweet.retweet().user())
            
            if tweet.retweet().is_quote():
                if tweet.retweet().quote():
                    self.record(tweet.retweet().quote())
                    self.record(tweet.retweet().quote().user())
                
        if tweet.is_quote():
            if tweet.quote():
                self.record(tweet.quote())
                self.record(tweet.quote().user())

        if self.length() >= self.N:
            self.write()
        
    def on_error(self, status_code):
        print("Error")
        print(status_code)
        return False


def recording(config, chunksize, languages, tracks):

    recorder = Recorder(chunksize)
    auth = tweepy.OAuthHandler(config['consumer_key'], config['consumer_secret'])
    auth.set_access_token(config['access_token'], config['access_token_secret'])
    api = tweepy.API(auth)
    stream = tweepy.Stream(auth = api.auth, listener=recorder)

    def shutdown():
        print(colored("Stop recording", "green"))
        recorder.write()
        stream.disconnect()

    atexit.register(shutdown)

    while True:
        try:                        
            stream = tweepy.Stream(auth = api.auth, listener=recorder)
            stream.filter(track=tracks, languages=languages)
        except Exception as ex:
            print(colored("Stream broken. Retrying ...", "red"))
            print(colored(ex, "red"))
            print(colored(traceback.format_exc(), "red"))
            time.sleep(10)


from setuptools import setup, find_packages
import twista

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name=twista.name,
    version=twista.VERSION,
    url='https://github.com/nkratzke/twista',
    license='MIT',
    author='Nane Kratzke',
    author_email='nane.kratzke@th-luebeck.de',
    description='Twitter streaming and graph-based analysis framework',
    long_description_content_type="text/markdown",
    long_description=long_description,
    python_requires=">=3.6,<4.0",
    scripts=[
        'bin/twista'
    ],
    packages=find_packages(),
    install_requires=[
        "certifi",
        "click             >= 7.0,  < 8.0",
        "tweepy            >= 3.5,  < 4.0",
        "python-dateutil   >= 2.6,  < 3.0",
        "tqdm              >= 4.35, < 5.0",
        "neo4j             >= 1.7,  < 2.0",
        "termcolor         >= 1.1,  < 1.2",
        "jupyterlab        >= 1.1,  < 2.0",
        "matplotlib        >= 3.1,  < 4.0",
        "py2neo            >= 4.3,  < 5.0"
    ]
)
